
#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <mmm/net_lib.h>
#include <mmm/dynamic.h>
#include <sys/wait.h>
#include <termios.h>
#include <fcntl.h>
#include <sys/time.h>
#include <signal.h>
#include "reg.h"
#include "talker/config.h"
#include "interface.h"

#define MAX_INP	2048
#define MAX_HISTORY	64
#define MAX_OUT	4096

#define	AUTO_LOGIN_RETRY	3

struct termios oldterm;
int cols=80;

char *output_filter=NULL;
char *output_filter_file=NULL;
char out_buf[MAX_OUT+1];
char prompt[128];
int prompt_len= -1;
int prompt_typ=PROMPT_NORMAL;
int login_id=0;

int restart( void );

void set_termios( int fd )
{ struct termios newterm;
	tcgetattr(fd,&oldterm);
	newterm=oldterm;
	newterm.c_iflag=(IGNBRK|IGNPAR);
	newterm.c_oflag=(0);
	newterm.c_cflag=(CS8|CLOCAL);
	newterm.c_lflag=(0);
	tcsetattr(fd,TCSANOW,&newterm);
}

void unset_termios( int fd )
{
	tcsetattr(fd,TCSANOW,&oldterm);
}


void per_lines( const char *s, void (*fn)(char*) )
{ char b[strlen(s)+1];
  int n;
	if( *s=='\r' )	s++;
  	while( *s!=0 )
	{	n=0;
		while( s[n]!=0 && s[n]!='\n' )
			n++;
		strncpy(b,s,n);	b[n]=0;
		fn(b);
		s+=n;
		if( *s!=0 )	s++;
		if( *s=='\r' )	s++;
	}
}

char inp_buf[MAX_INP+1];
int inp_offs=0;
int inp_pos=0;
int inp_cur=0;
int is_pager=0;
int is_pass=0;
char u_login[MAX_INP+1];
char u_pass[MAX_INP+1];


char *history[MAX_HISTORY];
int hist=0;
int hist_cur=0;
char *rep_cmd=NULL;

void add_history( void )
{
	inp_buf[inp_pos]=0;
	if( hist>=MAX_HISTORY-1 )
	{	free(history[0]);
		memmove(history,history+1,sizeof(history[0])*MAX_HISTORY-1);
		history[hist]=strdup(inp_buf);	/* stacilo by prazdne */
	}
	else
		history[hist++]=strdup(inp_buf);	/* deto */
	hist_cur=hist-1;
}

static int pl=1024,pofs=-1,ppos=-1;

void put_history( void )
{
	if( history[hist_cur]!=NULL )
		free(history[hist_cur]);
	inp_buf[inp_pos]=0;
	history[hist_cur]=strdup(inp_buf);
}

void get_history( int n )
{
	put_history();
	if( n<0 || n>=hist )
		return;
	strcpy(inp_buf,history[n]);
	inp_offs=0;
	inp_pos=inp_cur=strlen(inp_buf);
	hist_cur=n;
	pl=1024; pofs=-1; ppos=-1;
}

void display_spaces( int n )
{ static char *b="                                                                                ";
	while( n>80 )
	{	write(1,b,80);
		n-=80;
	}
	write(1,b,n);
}

void display_pass( int n )
{ static char *b="********************************************************************************";
	while( n>80 )
	{	write(1,b,80);
		n-=80;
	}
	write(1,b,n);
}

void display_input( void )
{ int l,lc;
  int acols=cols;
#ifndef INPUT_SIMPLE
  char buf[cols+16],c;
  int bp,i;

zaciatok:
	acols=cols;
	lc= -1;
	bp=0;
	for(l=0,i=inp_offs;i<inp_pos && l<=cols+1;i++)
	{	c=inp_buf[i];
		if( i==inp_cur )	lc=l;
		if( isprint(c) )
			buf[l++]=c;
		else if( c>=0 && c<' ' )
		{	buf[l++]='^';
			buf[l++]=c+'@';
		}
		else
		{	buf[l++]='\\';
			buf[l++]='x';
			buf[l++]="0123456789abcdef"[(c>>4)&0x0f];
			buf[l++]="0123456789abcdef"[c&0x0f];
		}
	}
	if( lc<0 )
	{	if( inp_cur > inp_offs )
			lc=l;
		else	lc=0;
	}
	buf[l]=0;
#endif
	if( is_pager )	return;
	if( prompt[0]!=0 )
	{	if( prompt_len<=0 )
			acols-=strlen(prompt);
		else	acols-=prompt_len;
	}
#ifdef INPUT_SIMPLE
	if( inp_cur-inp_offs >= acols )
		inp_offs=inp_cur-acols+acols/3;
	if( inp_cur<inp_offs+acols/3 )
	{	inp_offs=inp_cur-acols/3;
		if( inp_offs<0 )
			inp_offs=0;
	}
	l=inp_pos-inp_offs;
	lc=inp_cur-inp_offs;
#else
	if( lc >= acols )
	{	bp=lc-acols+acols/3;
		inp_offs+=bp;
		goto zaciatok;
//		l-=bp; lc-=bp;
	}
	if( lc < acols/3 && inp_offs>0 )
	{	inp_offs+=lc-acols/3;
		if( inp_offs<0 )
			inp_offs=0;
		goto zaciatok;
	}
#endif
	if( l>=acols )	l=acols-1;
	if( l!=pl || pofs!=inp_offs || ppos!=inp_pos )
	{	write(1,"\r",1);
		if( prompt[0]!=0 )
			write(1,prompt,strlen(prompt));
		if( is_pass )	display_pass(l);
#ifdef INPUT_SIMPLE
		else		write(1,inp_buf+inp_offs,l);
#else
		else		write(1,buf+bp,l);
#endif
		display_spaces(acols-1-l);
		pl=l; pofs=inp_offs; ppos=inp_pos;
	}
	write(1,"\r",1);
	if( prompt[0]!=0 )
		write(1,prompt,strlen(prompt));
	if( is_pass )	display_pass(lc);
#ifdef INPUT_SIMPLE
	else		write(1,inp_buf+inp_offs,lc);
#else
	else		write(1,buf+bp,lc);
#endif
}

void undisplay_input( void )
{
	write(1,"\r",1);
	display_spaces(cols-1);
	write(1,"\r",1);
	pl=1024; pofs=-1; ppos=-1;
}

int input( char c )
{ int l;
  static int stav=0,tab=0;
	if( c!='\t' )
		tab=0;
	if( stav== -1 )
	{	stav=0;
		// if( c=='\r' )	c='\n';
		goto raw;
	}
	if( c=='\n' || c=='\r' )
	{	if( is_pager )
		{	display("\n");
			is_pager=0;
			return(0);
		}
		l=inp_pos;
		if( !is_pass )
		{	hist_cur=hist-1;
			if( inp_pos>0 )
			{	put_history();
				add_history();
			}
		}
		else	is_pass=0;
		inp_cur=inp_pos=inp_offs=0;
		display_input();
		return(l);
	}
	else if( c=='Y'-'@' )
	{	if( !is_pass )
		{	hist_cur=hist-1;
			if( inp_pos>0 )
			{	put_history();
				add_history();
			}
		}
		inp_cur=inp_pos=inp_offs=0;
		display_input();
		return(-1);
	}
	else if( c=='X'-'@' )
	{	if( rep_cmd!=NULL )
		{	free(rep_cmd);
			rep_cmd=NULL;
		}
		else if( !is_pass )
		{	hist_cur=hist-1;
			if( inp_pos>0 )
			{	put_history();
				add_history();
			}
			inp_buf[inp_pos]=0;
			if( inp_pos>0 )
				rep_cmd=strdup(inp_buf);
		}
		inp_cur=inp_pos=inp_offs=0;
		display_input();
		return(-1);
	}
	else if( c=='Z'-'@' )
	{	if( !is_pass )
		{	hist_cur=hist-1;
			if( inp_pos>0 )
			{	put_history();
				add_history();
			}
		}
		inp_buf[inp_pos]=0;
		if( !strcmp(inp_buf,"quit") )
		{	
			unset_termios(0);
			exit(0);
		}
		else if( !strcmp(inp_buf,"restart") )
			restart();
		else if( !strcmp(inp_buf,"autologin") )
		{	if( auto_login == -1 )
			{	auto_login = 2;
				display("# Autologin is on\n");
			}
			else if( auto_login == 2 )
			{	auto_login = -1;
				display("# Autologin is off\n");
			}
			else
				display("# Autologin is ?\n");
		}
		else if( !strcmp(inp_buf,"help") )
		{	display("# Commands\n");
			display("# quit - quit talker\n");
			display("# restart - restart this client\n");
			display("# autologin - switch autologin on/off\n");
			display("# help - this help\n");
		}
		else	display("# Unknown command\n");
		inp_cur=inp_pos=inp_offs=0;
		display_input();
		return(-1);
	}
	else if( c=='U'-'@' )
	{	inp_cur=inp_pos=inp_offs=0;
		display_input();
		return(-1);
	}
	else if( c=='A'-'@' )
	{	inp_cur=0;
		display_input();
		return(-1);
	}
	else if( c=='E'-'@' )
	{	inp_cur=inp_pos;
		display_input();
		return(-1);
	}
	else if( c=='H'-'@' || c==0x7f )
	{	if( inp_cur>0 )
		{	inp_cur--;
			inp_pos--;
			if( inp_cur!=inp_pos )
				memmove(inp_buf+inp_cur,
					inp_buf+inp_cur+1,inp_pos-inp_cur);
		}
		display_input();
		return(-1);
	}
	else if( c=='P'-'@' )
	{	is_pass=1-is_pass;
		display_input();
		return(-1);
	}
	else if( c=='\t' )
	{	inp_buf[inp_pos]=0;
		if( tab )	inp_cur=tab_do(inp_buf,-1,MAX_INP);
		else		inp_cur=tab_do(inp_buf,inp_cur,MAX_INP);
		tab=1;
		inp_pos=strlen(inp_buf);
		display_input();
		return(-1);
	}
	else if( c=='V'-'@' )
	{	stav= -1;
		display_input();
		return(-1);
	}
	if( c=='['-'@' )
	{	if( stav==0 )	{ stav=1; return(-1); }
		else		stav=0;
	}
	if( stav==1 && (c=='O' || c=='['))	{ stav=2; return(-1); }
	if( stav>0 )
	{	switch( c )
		{	case 'A': /*UP*/
					get_history(hist_cur-1);
					break;
			case 'B': /*DOWN*/
					get_history(hist_cur+1);
					break;
			case 'C': /*RIGHT*/
					if( inp_cur<inp_pos )
						inp_cur++;
					break;
			case 'D': /*LEFT*/
					if( inp_cur>0 )
						inp_cur--;
					break;
		}
		display_input();
		stav=0;
		return(-1);
	}
raw:	if( inp_pos>=MAX_INP )	return(-2);
	if( inp_cur!=inp_pos )
		memmove(inp_buf+inp_cur+1,inp_buf+inp_cur,inp_pos-inp_cur);
	inp_buf[inp_cur++]=c;
	inp_pos++;
	display_input();
	return(-1);
}

void filter( char *s )
{ int i;
	i=0;
	while( s[i] )
	{	
//		if( s[i]==27 && s[i+1]!='[' )
//				s[i]='';
//		else
		if( s[i]!='\n' && s[i]!='\r' && s[i]!='\t'
			&& s[i]!='\a' && s[i]!=27
			&& !isprint(s[i]) )
		{	i++;
			continue;
		}
		*s=s[i];
		s++;
	}
	*s=0;
}

void display( char *ins )
{ int l;
  static char s[8192];
  static int sl=0;
  int ret=0,ll;
  char *p;
	if( *ins==0 )
	{	if( sl==0 )
			return;
		ins="\n\r";
	}

	if( (p=strrchr(ins,'\n'))==NULL )
	{	ll=strlen(ins);
		ret=1;
	}
	else
	{	if( p[1]=='\r' )
			p++;
		ll=p-ins+1;
	}
	if( ll>8191 )
	{	ins+= ll-8191;
		ll= 8191;
	}
	if( ll>=8191-sl )
	{	memmove(s,s+sl-(8191-ll),8191-ll);
			sl=8191-ll;
	}
	strncpy(s+sl,ins,ll);
	ins+=ll;
	sl+=ll;
	s[sl]=0;

	if( is_pager )
		is_pager=0;
	else
		undisplay_input();

	if( ret )
		goto tprom;

	// l=strlen(s);
	l= sl;
	if( output_filter )
	{ int p[2];
		pipe(p);
		switch( fork() )
		{ darg_t *da;
			case 0:
				close(0);
				dup(p[0]);
				close(p[0]);
				close(p[1]);
				da=darg_line2argv(output_filter);
				execvp(darg_argv(da)[0],darg_argv(da));
				exit(-1);
			case -1:
				close(p[0]);
				close(p[1]);
				break;
			default:
				close(p[0]);
				write(p[1],s,l);
				close(p[1]);
				wait(NULL);
				break;
		}
	}
	else if( output_filter_file )
	{ char *p;
		p=sed_file(s,output_filter_file);
		write(1,p,strlen(p));
		free(p);
	}
	else
		write(1,s,l);
	
	sl=0;
	l=strlen(ins);
	if( l>8191 )
	{	ins+= l - 8191;
		l=8181;
	}
	strncpy(s,ins,l);
	sl=l;
	s[sl]=0;
tprom:
	if( test_prompt && sl>0 && sl<80 )
	{ char *p;
	  regmatch_t m[16];
		p=reg_subst(s,"\\[[0-9]+(;[0-9]+)*m","","g");
		if( *p=='\r' )	p++;
//printf("ZZZ: testujem : (%d) %s\n",strlen(p),p);
		if( talker_prompt!=NULL && !strreg(p,talker_prompt,m,16) )
		{	strcpy(prompt,s);
			prompt_len=strlen(p);
			prompt_typ=PROMPT_NORMAL;
			sl=0;
			s[sl]=0;
//printf("ZZZ: mam\n");
		}
		else if( pager_prompt!=NULL && !strreg(p,pager_prompt,m,16) )
		{	is_pager=1;
			write(1,s,sl);
			sl=0;
			s[sl]=0;
		}
		else if( login_prompt!=NULL && !strreg(p,login_prompt,m,16) )
		{	strcpy(prompt,s);
			prompt_len=strlen(p);
			prompt_typ=PROMPT_LOGIN;
			sl=0;
			s[sl]=0;
		}
		else if( pass_prompt!=NULL && !strreg(p,pass_prompt,m,16) )
		{	strcpy(prompt,s);
			prompt_len=strlen(p);
			prompt_typ=PROMPT_PASS;
			is_pass=1;
			sl=0;
			s[sl]=0;
		}
		else if( prompt_typ!=PROMPT_NORMAL )
		{	prompt_typ=PROMPT_NORMAL;
			prompt[0]=0;
			prompt_len=0;
		}
		free(p);
	}
	else if( sl==0 && prompt_typ!=PROMPT_NORMAL )
	{	prompt_typ=PROMPT_NORMAL;
		prompt[0]=0;
		prompt_len=0;
	}

/*
	if( !is_pager )
		if( s[strlen(s)-1]!='\n' )
			write(1,"\n",1);
*/
	display_input();
}

char *g_av[8];
char g_sock[16];

int restart( void )
{
	undisplay_input();
	execvp(g_av[0],g_av);
	display("# Restart error\n");
	return(-1);
}

void sigsegv_handler( int signum )
{
	output_filter=NULL;
	output_filter_file=NULL;
	display("# Segmentation fault\n");
	restart();
	unset_termios(0);
	exit(-1);
}

void sigusr1_handler( int signum )
{
	if( auto_login==2 )
		auto_login= -1;
	else
		auto_login= 0;
}

int main( int argc, char *argv[] )
{ int i,l,a,gn;
  unsigned int ip;
  int s= -1;
  fd_set rd;
  char c,*p;
  struct timeval tv;
  int auto_login_retry=AUTO_LOGIN_RETRY;
	gn=0;
	g_av[gn++]=argv[0];
  	prompt[0]=0;
	prompt_len= -1;
	p=getenv("COLUMNS");
	if( p!=NULL && (i=atoi(p))>0 )
		cols=i;
	for(a=1;a<argc;a++)
	{	if( argv[a][0]=='-' )
		{	if( argv[a][1]=='o' && a+1<argc )
			{	g_av[gn++]=argv[a];
				a++;
				g_av[gn++]=argv[a];
				output_filter=argv[a];
			}
			else if( argv[a][1]=='f' && a+1<argc )
			{	g_av[gn++]=argv[a];
				a++;
				g_av[gn++]=argv[a];
				output_filter_file=argv[a];
			}
			else if( argv[a][1]=='#' && a+1<argc )
			{	a++;
				s=atoi(argv[a]);
			}
		}
	}

relogin:
	if( s<0 )
	{
		net_name2ip(TALKER,&ip);
		s=net_connect(ip,TALKER_PORT);
		if( talker_initstring!=NULL )
			write(s,talker_initstring,strlen(talker_initstring));
		login_id++;
	}
	else
		display("# Restarted\n");
	fcntl(s,F_SETFD,0);	/* ! FD_CLOEXEC */
	g_av[gn++]="-#";
	sprintf(g_sock,"%d",s);
	g_av[gn++]=g_sock;
	g_av[gn]=NULL;

	signal(SIGSEGV,sigsegv_handler);
	signal(SIGUSR1,sigusr1_handler);
	set_termios(0);
	add_history();
	while( s>=0 )
	{	FD_ZERO(&rd);
		FD_SET(0,&rd);
		FD_SET(s,&rd);
		tv.tv_sec=1*60;
		tv.tv_usec=0;
		select(s+1,&rd,NULL,NULL,&tv);
		if( FD_ISSET(0,&rd) )
		{	if( read(0,&c,1)<=0 )
				break;
			if( (l=input(c))>=0 )
			{	if( prompt_typ==PROMPT_LOGIN )
				{	strncpy(u_login,inp_buf,l);
					u_login[l]=0;
				} else if( prompt_typ==PROMPT_PASS )
				{	strncpy(u_pass,inp_buf,l);
					u_pass[l]=0;
					if( auto_login==1 )
						auto_login=2;
				}
				else
					tab_add_kb(inp_buf);
				write(s,inp_buf,l);
				write(s,"\n",1);
			}
			if( inp_pos==0 && rep_cmd!=NULL )
			{	strcpy(inp_buf,rep_cmd);
				inp_cur=inp_pos=strlen(inp_buf);
				inp_offs=0;
				pl=1024; pofs=-1; ppos=-1;
				display_input();
			}
		}
		if( FD_ISSET(s,&rd) )
		{	if( (l=read(s,out_buf,MAX_OUT))<=0 )
				break;
			out_buf[l]=0;
			filter(out_buf);
			per_lines(out_buf,tab_add);
			p=ai(out_buf,s);
			display(p);
			if( auto_login>=3 )
			{	if( auto_login==3 && prompt_typ==PROMPT_LOGIN )
				{	// sleep(1);
					write(s,u_login,strlen(u_login));
					write(s,"\n",1);
					auto_login=4;
				}
				else if( auto_login==4 && prompt_typ==PROMPT_PASS )
				{	// sleep(1);
					write(s,u_pass,strlen(u_pass));
					write(s,"\n",1);
					is_pass=0;
					auto_login=2;
				}
			}
			auto_login_retry=AUTO_LOGIN_RETRY;
   			if( p!=out_buf )
   				free(p);
//			display(out_buf);
		}
		else if( ! FD_ISSET(0,&rd) )
		{	if( is_pager )
			{	display("\n");
				is_pager=0;
				write(s,"\n",1);
			}
		}

	}
	if( s>=0 )
		close(s);
	s= -1;
	if( auto_login>=2 && auto_login_retry>0 )
	{	auto_login=3;
		auto_login_retry--;
		goto relogin;
	}
	unset_termios(0);
	return(0);
}
